/*
 *   Copyright 2014 OSBI Ltd
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */
package org.saiku.service.util.export;

import org.saiku.olap.util.SaikuProperties;

import java.io.IOException;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.Clob;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;

public class ResultSetHelper {

  private final NumberFormat numberFormat;
  private final SimpleDateFormat dateFormat;
  private final SimpleDateFormat timestampFormat;

  public ResultSetHelper() {
    this.numberFormat = NumberFormat.getInstance( SaikuProperties.locale );
    if ( SaikuProperties.webExportCsvUseFormattedValue ) {
      numberFormat.setGroupingUsed( true );
    } else {
      numberFormat.setGroupingUsed( false );
    }
    this.dateFormat = new SimpleDateFormat( SaikuProperties.webExportCsvDateFormat );
    this.timestampFormat = new SimpleDateFormat( SaikuProperties.webExportCsvTimestampFormat );
  }

  private String formatNumber( Object number ) {
    return numberFormat.format( number );
  }

  private String read( Clob c ) throws SQLException, IOException {
    StringBuilder sb = new StringBuilder( (int) c.length() );
    Reader r = c.getCharacterStream();
    char[] cbuf = new char[ 2048 ];
    int n = 0;
    while ( ( n = r.read( cbuf, 0, cbuf.length ) ) != -1 ) {
      if ( n > 0 ) {
        sb.append( cbuf, 0, n );
      }
    }
    return sb.toString();
  }

  public String getValue( ResultSet rs, int colType, int colIndex ) throws SQLException, IOException {
    String value = "";

    switch( colType ) {
      case Types.BIT:
        Object bit = rs.getObject( colIndex );
        if ( bit != null ) {
          value = String.valueOf( bit );
        }
        break;
      case Types.BOOLEAN:
        boolean b = rs.getBoolean( colIndex );
        if ( !rs.wasNull() ) {
          value = Boolean.valueOf( b ).toString();
        }
        break;
      case Types.CLOB:
        Clob c = rs.getClob( colIndex );
        if ( c != null ) {
          value = this.read( c );
        }
        break;
      case Types.BIGINT:
      case Types.DECIMAL:
      case Types.DOUBLE:
      case Types.FLOAT:
      case Types.REAL:
      case Types.NUMERIC:
        BigDecimal bd = rs.getBigDecimal( colIndex );
        if ( bd != null ) {
          value = this.formatNumber( bd ); // value = "" + bd.doubleValue();
        }
        break;
      case Types.INTEGER:
      case Types.TINYINT:
      case Types.SMALLINT:
        int intValue = rs.getInt( colIndex );
        if ( !rs.wasNull() ) {
          value = this.formatNumber( intValue ); // value = "" + intValue;
        }
        break;
      case Types.JAVA_OBJECT:
        Object obj = rs.getObject( colIndex );
        if ( obj != null ) {
          value = String.valueOf( obj );
        }
        break;
      case Types.DATE:
        java.sql.Date date = rs.getDate( colIndex );
        if ( date != null ) {
          value = this.dateFormat.format( date );
        }
        break;
      case Types.TIME:
        Time t = rs.getTime( colIndex );
        if ( t != null ) {
          value = t.toString();
        }
        break;
      case Types.TIMESTAMP:
        Timestamp tstamp = rs.getTimestamp( colIndex );
        if ( tstamp != null ) {
          value = this.timestampFormat.format( tstamp );
        }
        break;
      case Types.LONGVARCHAR:
      case Types.VARCHAR:
      case Types.CHAR:
        value = rs.getString( colIndex );
        break;
      default:
        value = "";
    }

    if ( value == null ) {
      value = "";
    }

    return value;

  }
}
